import csv
import io
import json
import pprint

from pygments.lexers.special import TextLexer
from tabulate import tabulate
from pygments.lexers.data import JsonLexer
from pygments.lexers.html import HtmlLexer
from pygments.lexers.python import PythonLexer

from textomatic.model import ProcessedCommand
from textomatic.processor.common import run_jq
from textomatic.processor.registry import Registry

DEFAULT_LEXER = PythonLexer


class Output:
    lexer = DEFAULT_LEXER

    def create_output(self, rows, processed_command: ProcessedCommand) -> str:
        raise NotImplementedError


class PythonLiteralOutput(Output):
    lexer = PythonLexer

    def create_output(self, rows, processed_command: ProcessedCommand) -> str:
        return pprint.pformat(
            rows,
            indent=1,
            compact=False,
            sort_dicts=False,
        )


class JsonOutput(Output):
    lexer = JsonLexer

    def create_output(self, rows, processed_command: ProcessedCommand) -> str:
        return json.dumps(rows, indent=4)


class JsonLinesOutput(Output):
    lexer = JsonLexer

    def create_output(self, rows, processed_command: ProcessedCommand) -> str:
        return "\n".join(json.dumps(r) for r in rows)


class CSVOutput(Output):
    lexer = JsonLexer

    class Dialect(csv.Dialect):
        delimiter = ","
        doublequote = True
        escapechar = "\\"
        lineterminator = "\n"
        quotechar = '"'
        quoting = csv.QUOTE_ALL
        skipinitialspace = True
        strict = False

    dialect = Dialect()

    def create_output(self, rows, processed_command: ProcessedCommand) -> str:
        out = io.StringIO()
        writer = csv.writer(out, self.dialect)
        if processed_command.headers:
            writer.writerow(processed_command.headers.values())
        writer.writerows(rows)
        return out.getvalue()


class TableOutput(Output):
    lexer = JsonLexer

    def create_output(self, rows, processed_command: ProcessedCommand) -> str:
        kwargs = {}
        if processed_command.headers:
            kwargs["headers"] = processed_command.headers.values()
        return tabulate(rows, tablefmt="fancy_grid", **kwargs)


class HTMLOutput(Output):
    lexer = HtmlLexer

    def create_output(self, rows, processed_command: ProcessedCommand) -> str:
        kwargs = {}
        if processed_command.headers:
            kwargs["headers"] = processed_command.headers.values()
        return tabulate(rows, tablefmt="html", **kwargs)


class JQOutput(Output):
    lexer = JsonLexer

    def __init__(self, args):
        self.args = args

    def create_output(self, rows, processed_command: ProcessedCommand) -> str:
        return run_jq(rows, self.args)


class NopOutput(Output):
    lexer = TextLexer

    def create_output(self, rows, processed_command: ProcessedCommand) -> str:
        return rows


registry = Registry(
    attr="outputs",
    tpe=Output,
    default_alias="l",
    data={
        "l": PythonLiteralOutput(),
        "j": JsonOutput(),
        "jl": JsonLinesOutput(),
        "c": CSVOutput(),
        "t": TableOutput(),
        "h": HTMLOutput(),
        "jq": JQOutput,
        "n": NopOutput(),
    },
)



import ast
import json

from textomatic.context import ProcessContext
from textomatic.processor import parser, outputs, inputs, macros
from textomatic.exceptions import ProcessException
from textomatic.model import ProcessedInput, ProcessedCommand, NO_DEFAULT, MISSING

DEFAULT_CMD = ProcessedCommand("")


def get(attr, _):
    def fn(v):
        result = v[attr]
        if result is MISSING:
            raise IndexError(f"Missing {attr}")
        return result

    return fn


def getsafe(attr, default=None):
    def fn(v):
        try:
            result = v[attr]
            if result is MISSING:
                result = default
            return result
        except Exception:
            return default

    return fn


def process(text: str, cmd: str, ctx: ProcessContext, trigger: str = None):
    processed_cmd, changed = _process_cmd(ctx, cmd)
    input_objs = inputs.registry.get(processed_cmd)
    output_objs = outputs.registry.get(processed_cmd)

    if trigger == "cmd" and not changed:
        return None

    if trigger != "cmd" or any(attr in changed for attr in ["delimiter", "inputs", "has_header", "raw"]):
        rows, headers_list = text, []
        for input_obj in input_objs:
            prev_headers = headers_list
            rows, headers_list = input_obj.get_rows(rows, processed_cmd)
            headers_list = headers_list or prev_headers
        headers = {i: h for i, h in enumerate(headers_list)}
        ctx.processed_input = ProcessedInput(rows, headers)
        if headers:
            processed_cmd.has_header = True
    else:
        headers = ctx.processed_input.headers
        rows = ctx.processed_input.rows
    rows = _process_rows(processed_cmd, headers, rows)
    processed_cmd.headers = _extract_output_headers(processed_cmd.structure, headers)

    result = rows
    for output_obj in output_objs:
        result = output_obj.create_output(result, processed_cmd)
    return result


def _extract_output_headers(structure: parser.StructureData, input_headers):
    if not structure or not structure.fields:
        return input_headers
    result = {}
    for i, field in enumerate(structure.fields):
        result[i] = _extract_structure_field_name(field, input_headers)
    return result


def _extract_structure_field_name(field: parser.ParseData, headers):
    if isinstance(field, parser.KeyToValueData):
        name = field.key.value
    elif isinstance(field, parser.RefData):
        first, rest = field.value[0], field.value[1:]
        if isinstance(first, parser.LocData):
            num = first.value
            if num > 0:
                num -= 1
            else:
                num += len(headers)
            if num in headers:
                first_name = headers[num]
            else:
                first_name = str(first.value)
        elif isinstance(first, parser.IdData):
            first_name = first.value
        else:
            raise ProcessException(f"Unexpected first field: {first}")
        name = ".".join([first_name] + [str(f.value) for f in rest])
    else:
        raise ProcessException(f"Unexpected field: {field}")
    return name


def _process_rows(processed_cmd, headers, rows):
    if processed_cmd.raw:
        return rows
    headers_inverse = {h: i for i, h in headers.items()}
    type_processors = _build_row_types_processor(processed_cmd.types, headers_inverse)
    structure_processor = _build_row_structure_processor(processed_cmd.structure, headers, headers_inverse)
    return [_process_row(type_processors, structure_processor, r) for r in rows]


def _process_row(type_processors, structure_processor, row):
    if type_processors:
        row = row[:]
        for i, type_processor, optional_ref, optional_type, default in type_processors:
            try:
                value = row[i]
                if value is MISSING:
                    if optional_ref:
                        row[i] = default
                        continue
                    else:
                        raise IndexError(f"Missing index {i}")
                try:
                    row[i] = type_processor(value)
                except ProcessException:
                    raise
                except Exception:
                    if not optional_type:
                        raise
                    row[i] = default
            except IndexError:
                if not optional_ref:
                    raise
    return structure_processor(row)


def _build_row_types_processor(types, headers_inverse):
    type_processors = []
    if not types:
        return type_processors
    saw_header = False
    for i, col_type in enumerate(types):
        if isinstance(col_type, parser.TypeDefData):
            if saw_header:
                raise ProcessException("Cannot specify types for indexed columns after named columns")
            type_def = col_type
            optional_type = type_def.optional
            optional_ref = optional_type
            default = type_def.default
        elif isinstance(col_type, parser.KeyToValueData):
            saw_header = True
            type_def = col_type.value
            optional_type = type_def.optional
            default = type_def.default
            if isinstance(col_type.key, parser.LocData):
                optional_ref = col_type.key.optional
                i = col_type.key.value
                if i > 0:
                    i -= 1
            elif isinstance(col_type.key, parser.IdData):
                optional_ref = col_type.key.optional
                header = col_type.key.value
                i = headers_inverse[header]
            else:
                raise ProcessException(f"Unknown key type: {col_type.key}")
        else:
            raise ProcessException(f"Unknown col_type: {col_type}")
        row_type_processor = _build_type_processor(type_def.type)
        if default is not NO_DEFAULT and not (optional_ref or optional_type):
            optional_ref = True
            optional_type = True
        if default is NO_DEFAULT:
            default = None
        else:
            try:
                default = ast.literal_eval(default)
            except Exception as e:
                raise ProcessException(f"Invalid default literal value: {default}") from e
        type_processors.append((i, row_type_processor, optional_ref, optional_type, default))
    return type_processors


def _build_type_processor(t):
    if t in {"_", "s"}:
        return str
    elif t == "f":
        return float
    elif t == "i":
        return int
    elif t == "b":
        return lambda v: str(v).lower() in {"true", "yes", "y", "on", "1"}
    elif t == "j":
        return lambda v: json.loads(str(v))
    elif t == "l":
        return lambda v: ast.literal_eval(str(v))
    elif t == "d":
        raise ProcessException("Dates are not supported yet")
    else:
        raise ProcessException(f"Unsupported column type: {t}")


def _build_row_structure_processor(structure, headers, headers_inverse):
    def replace_missing_from_row(r):
        return [replace_missing(v) for v in r]

    def replace_missing(v):
        return None if v is MISSING else v

    if not structure:
        return lambda r: replace_missing_from_row(r)

    def impl(data: parser.ParseData, values):
        if isinstance(data, parser.StructureData):
            assert values
            if not data.fields:
                if data.type == "[]":
                    return lambda r: replace_missing_from_row(r)
                elif data.type == "()":
                    return lambda r: tuple(replace_missing(i) for i in r)
                elif data.type == "s()":
                    return lambda r: set(replace_missing(i) for i in r)
                elif data.type in {"{}", "d()"}:
                    return lambda r: dict(zip(headers.values(), replace_missing_from_row(r)))
                else:
                    raise ProcessException(f"Unsupported data type: {data.type}")
            else:
                values = [impl(f, values=True) for f in data.fields]
                if data.type == "[]":
                    return lambda row: [v(row) for v in values]
                elif data.type == "()":
                    return lambda row: tuple(v(row) for v in values)
                elif data.type == "s()":
                    return lambda row: set(v(row) for v in values)
                elif data.type in {"{}", "d()"}:
                    keys = [impl(f, values=False) for f in data.fields]
                    return lambda row: dict(zip(keys, [v(row) for v in values]))
                else:
                    raise ProcessException(f"Unsupported data type: {data.type}")
        elif isinstance(data, parser.KeyToValueData):
            if values:
                return impl(data.value, values=True)
            else:
                return data.key.value
        elif isinstance(data, parser.RefData):
            if values:
                path = data.value
                default = data.default
                has_default = default is not NO_DEFAULT
                if has_default:
                    try:
                        default = ast.literal_eval(default)
                    except Exception as e:
                        raise ProcessException(f"Invalid default literal value: {default}") from e
                else:
                    default = None
                has_optional = any(
                    (isinstance(p, parser.IdData) and p.optional) or (isinstance(p, parser.LocData) and p.optional)
                    for p in path
                )
                value_processor = []
                for part in path:
                    getter = get
                    if isinstance(part, parser.IdData):
                        if part.optional or (has_default and not has_optional):
                            getter = getsafe
                        key = part.value
                        if not value_processor:
                            index = headers_inverse.get(key)
                            value_processor.append(getter(index, default))
                        else:
                            value_processor.append(getter(key, default))
                    elif isinstance(part, parser.LocData):
                        if part.optional or (has_default and not has_optional):
                            getter = getsafe
                        index = part.value
                        if not value_processor:
                            if index > 0:
                                index -= 1
                        value_processor.append(getter(index, default))
                    else:
                        raise ProcessException(f"Unexpected part {part}")

                def fn(row):
                    result = row
                    for p in value_processor:
                        result = p(result)
                        if result is default:
                            break
                    return result

                return fn
            else:
                return _extract_structure_field_name(data, headers)
        else:
            raise ProcessException(f"Unsupported parser data: {data}")

    return impl(structure, values=True)


def _process_cmd(ctx, cmd) -> ProcessedCommand:
    cmd = cmd.strip()

    if cmd.startswith("@"):
        alias, *args = cmd.split(" ", 1)
        alias = alias[1:]
        macro = macros.get(alias)
        if args:
            args = macro["split"](args[0].strip())
        cmd = macro["fn"](*args)

    changed = set()
    if cmd == ":" or (ctx.processed_command and ctx.processed_command.cmd == cmd):
        return ctx.processed_command, changed

    previous_command = ctx.processed_command
    result = ProcessedCommand(cmd)

    if cmd.startswith(":") and len(cmd) >= 2:
        cmd_separator = cmd[1]
        cmd = cmd[2:]
    else:
        cmd_separator = ";"
    for expression in cmd.split(cmd_separator):
        expression = expression.strip()
        if not expression:
            continue
        if expression == "h":
            result.has_header = True
            continue
        if expression == "r":
            result.raw = True
            continue
        expression_split = expression.split(":", 1)
        if len(expression_split) < 2:
            continue
        expression_type, expression_body = expression_split
        expression_type, expression_body = expression_type.strip(), expression_body.strip()
        if expression_type not in "dhtsio":
            raise ProcessException(f"Unsupported command type: {expression_type}")
        elif expression_type == "d":
            if expression_body.startswith("\\"):
                expression_body = ast.literal_eval(f'"{expression_body}"')
            result.delimiter = expression_body or DEFAULT_CMD.delimiter
        elif expression_type == "t":
            if expression_body:
                result.types = parser.parse_types(expression_body)
            else:
                result.types = DEFAULT_CMD.types
        elif expression_type == "s":
            if expression_body:
                result.structure = parser.parse_structure(expression_body)
            else:
                result.structure = DEFAULT_CMD.structure
        elif expression_type == "i":
            if expression_body:
                result.inputs = parser.parse_processors(expression_body)
            else:
                result.input = DEFAULT_CMD.inputs
        elif expression_type == "o":
            if expression_body:
                result.outputs = parser.parse_processors(expression_body)
            else:
                result.outputs = DEFAULT_CMD.outputs
        else:
            raise ProcessException("Unexpected")

    previous_command.cmd = cmd
    ctx.processed_command = result
    for attr in ["cmd", "delimiter", "outputs", "inputs", "structure", "types", "has_header", "raw"]:
        if getattr(previous_command, attr) != getattr(result, attr):
            changed.add(attr)
    return result, changed



from textwrap import dedent

from prompt_toolkit import Application, HTML
from prompt_toolkit.clipboard.pyperclip import PyperclipClipboard
from prompt_toolkit.enums import EditingMode
from prompt_toolkit.filters import Condition
from prompt_toolkit.layout import (
    Layout,
    HSplit,
    VSplit,
    Window,
    BufferControl,
    DynamicContainer,
    FloatContainer,
    Float,
    ConditionalContainer,
)
from prompt_toolkit.layout.processors import BeforeInput
from prompt_toolkit.lexers import PygmentsLexer, DynamicLexer
from prompt_toolkit.output import ColorDepth
from prompt_toolkit.widgets import FormattedTextToolbar, VerticalLine, Dialog, Label

from textomatic import context
from textomatic.app import style
from textomatic.app.lexer import CommandLexer
from textomatic.processor import outputs, inputs
from textomatic.exceptions import ProcessException
from textomatic.app.keys import kb, cmd_kb
from textomatic.processor.process import process
from textomatic.app.style import application_style
from textomatic.app.widgets import textbox
from textomatic.processor.registry import Registry
from textomatic.text import as_printable


class AppBuilder:
    def __init__(self):
        self.ctx = context.get()
        self.process_ctx = context.get_process()
        self.ctx.process_fn = self.process_key
        self.lexer_threshold = 10000
        self.input_lexer = ToggledLexer(inputs.registry, take_last=False)
        self.output_lexer = ToggledLexer(outputs.registry, take_last=True)

    def create_app(self, focus):
        ctx = self.ctx
        ctx.cmd_buffer.on_text_changed += self.process_key
        ctx.input_buffer.on_text_changed += self.process_key
        ctx.cmd_buffer.cursor_position = len(ctx.cmd_buffer.text)
        if focus:
            focused_element = {
                "COMMAND": ctx.cmd_buffer,
                "c": ctx.cmd_buffer,
                "INPUT": ctx.input_buffer,
                "i": ctx.input_buffer,
                "OUTPUT": ctx.output_buffer,
                "o": ctx.output_buffer,
            }[focus]
        else:
            focused_element = ctx.cmd_buffer if ctx.input_buffer.text else ctx.input_buffer
        layout = Layout(
            FloatContainer(
                HSplit(
                    [
                        self.create_boxes(),
                        self.create_status_bar(),
                        self.create_cmd_bar(),
                    ]
                ),
                floats=[
                    Float(self.create_help()),
                ],
            ),
            focused_element=focused_element,
        )
        return Application(
            key_bindings=kb,
            editing_mode=EditingMode.VI,
            clipboard=PyperclipClipboard(),
            color_depth=ColorDepth.from_env() or ColorDepth.DEPTH_24_BIT,
            full_screen=True,
            mouse_support=True,
            include_default_pygments_style=False,
            style=application_style,
            layout=layout,
        )

    def process_key(self, trigger=None):
        ctx = self.ctx
        ctx.current_error = None
        ctx.display_help = False
        self.input_lexer.enabled = len(ctx.input_buffer.text) < self.lexer_threshold
        if not ctx.live and trigger != "run":
            ctx.dirty = True
            return
        if trigger:
            trigger = "cmd" if trigger is ctx.cmd_buffer else "input"
        try:
            result = process(
                text=ctx.input_buffer.text,
                cmd=ctx.cmd_buffer.text,
                ctx=self.process_ctx,
                trigger=trigger,
            )
            if result is not None:
                if not isinstance(result, str):
                    result = str(result)
                self.output_lexer.enabled = len(result) < self.lexer_threshold
                ctx.output_buffer._set_text(result)
            ctx.dirty = False
        except Exception as e:
            if ctx.non_interactive:
                raise
            if isinstance(e, ProcessException):
                ctx.current_error = str(e)
            else:
                ctx.current_error = f"{e.__class__.__name__}: {e}"

    def create_status_bar(self):
        ctx = self.ctx
        process_ctx = self.process_ctx

        def _get_text():
            cmd = process_ctx.processed_command
            if ctx.current_error:
                result = ["{current_error}"]
            else:
                current = ctx.app.layout.current_buffer
                focused = {
                    id(ctx.cmd_buffer): "COMMAND",
                    id(ctx.input_buffer): "INPUT",
                    id(ctx.output_buffer): "OUTPUT",
                }[id(current)]
                mode = "live" if ctx.live else "manual"
                elem = "ansired" if ctx.dirty else "ansigreen"
                mode = f"<{elem}>{mode}</{elem}>"
                result = [
                    f"{ctx.app.vi_state.input_mode}|",
                    f"{focused}|",
                    f"{mode}|",
                    "in:{input}|",
                    "out:{output}|",
                    "delim:{delimiter}|",
                    f"header:{str(cmd.has_header).lower()}|",
                    f"raw:{str(cmd.raw).lower()}",
                ]
                if ctx.copied_to_clipboard:
                    result.append(" [Copied output to clipboard]")
            inp = ",".join([i.alias for i in cmd.inputs or []] or ["c"])
            out = ",".join([o.alias for o in cmd.outputs or []] or ["l"])
            return HTML("".join(result)).format(
                input=inp,
                output=out,
                delimiter=as_printable(cmd.delimiter) or "auto",
                current_error=str(ctx.current_error),
            )

        def _get_style():
            if ctx.copied_to_clipboard:
                cls = style.STATUS_BAR_NOTIFY
            elif ctx.current_error:
                cls = style.STATUS_BAR_ERROR
            else:
                cls = style.STATUS_BAR
            return f"class:{cls}"

        return VSplit([FormattedTextToolbar(_get_text)], style=_get_style)

    def create_boxes(self):
        ctx = self.ctx
        input_box = textbox(ctx.input_buffer, "INPUT", lexer=self.input_lexer)
        output_box = textbox(ctx.output_buffer, "OUTPUT", lexer=self.output_lexer)
        vertical_boxes = VSplit([input_box, VerticalLine(), output_box])
        horizontal_boxes = HSplit([input_box, output_box])
        return DynamicContainer(lambda: vertical_boxes if ctx.box_veritcal_orientation else horizontal_boxes)

    def create_cmd_bar(self):
        ctx = self.ctx

        def get_cmd_prefix():
            has_focus = ctx.app.layout.current_buffer is ctx.cmd_buffer
            cls = style.COMMAND_FOCUSED
            if has_focus:
                text = f"<{cls}>> </{cls}>"
            else:
                text = "> "
            return HTML(text)

        return Window(
            BufferControl(
                ctx.cmd_buffer,
                focus_on_click=True,
                input_processors=[BeforeInput(get_cmd_prefix)],
                key_bindings=cmd_kb,
                lexer=PygmentsLexer(CommandLexer),
            ),
            height=1,
        )

    def create_help(self):
        ctx = self.ctx
        dialog = Dialog(
            Label(
                dedent(
                    """
                    Tab               Switch focus between INPUT/OUTPUT/COMMAND
                    Ctrl-P            Copy OUTPUT to clipboard
                    Ctrl-O            Exit and print OUTPUT to stdout
                    Ctrl-C            Exit
                    Ctrl-T            Toggle vertical/horizontal view
                    Ctrl-L            Toggle live/manual mode
                    Ctrl-R            Run processing (when in manual mode)
                    Enter/Ctrl-M      Run processing (when in manual mode, in COMMAND)
                    F1                Toggle help
                """
                ).strip(),
                dont_extend_width=False,
                dont_extend_height=False,
            ),
            title="Help",
        )
        return ConditionalContainer(dialog, Condition(lambda: ctx.display_help))


class ToggledLexer(DynamicLexer):
    def __init__(self, registry: Registry, take_last):
        self.enabled = True
        self.registry = registry

        def get_lexer():
            if not self.enabled:
                return None
            process_ctx = context.get_process()
            processors = self.registry.get(process_ctx.processed_command, safe=True)
            processor = processors[-1] if take_last else processors[0]
            return PygmentsLexer(processor.lexer)

        super().__init__(get_lexer)


def process_one():
    builder = AppBuilder()
    builder.process_key()


def create_app(focus):
    builder = AppBuilder()
    builder.process_key()
    return builder.create_app(focus)



from textomatic.exceptions import ProcessException
from textomatic.model import ProcessedCommand


class Registry:
    def __init__(self, attr, tpe, default_alias, data=None, default_raw="n"):
        self.attr = attr
        self.type = tpe
        self.data = data or {}
        self.default_alias = default_alias
        self.default_raw = default_raw

    def register(self, alias):
        def registrar(value):
            assert (isinstance(value, type) and issubclass(value, self.type)) or isinstance(value, self.type)
            self.data[alias] = value
            return value

        return registrar

    def get(self, processed_cmd: ProcessedCommand, safe=False):
        if processed_cmd.raw:
            default = [self.data[self.default_raw]]
        else:
            default = [self.data[self.default_alias]]
        data_processors_configs = getattr(processed_cmd, self.attr)
        if not data_processors_configs:
            return default
        result = []
        for data_processor_config in data_processors_configs:
            alias = data_processor_config.alias
            args = data_processor_config.args
            try:
                obj_or_cls = self.data[alias]
            except KeyError:
                if safe:
                    return default
                raise ProcessException(f"Unregistered {self.type.__name__.lower()}: {alias}")
            if isinstance(obj_or_cls, type) and issubclass(obj_or_cls, self.type):
                result.append(obj_or_cls(args))
            else:
                result.append(obj_or_cls)
        return result



from pygments.lexer import RegexLexer, bygroups
from pygments.token import Text, Keyword, Operator


class CommandLexer(RegexLexer):
    tokens = {
        "root": [
            (r"([iosthrd])(:)", bygroups(Keyword, Text)),
            (r";", Operator),
            (r".", Text),
        ],
    }



import json
import shlex
from typing import List, Any, Mapping

import clevercsv
from pygments.lexers.data import JsonLexer
from pygments.lexers.python import PythonLexer
from pygments.lexers.special import TextLexer

from textomatic.model import ProcessedCommand, MISSING
from textomatic.processor.common import run_jq
from textomatic.processor.registry import Registry

DEFAULT_LEXER = PythonLexer


class Input:
    lexer = DEFAULT_LEXER

    def get_rows(self, text: str, processed_cmd: ProcessedCommand) -> (List[Any], Mapping[int, str]):
        raise NotImplementedError


class CSVInput(Input):
    lexer = PythonLexer

    def get_rows(self, text: str, processed_cmd: ProcessedCommand) -> (List[Any], Mapping[int, str]):
        headers_list = []
        delimiters = [processed_cmd.delimiter] if processed_cmd.delimiter else None
        dialect = clevercsv.Sniffer().sniff(text[:10000], delimiters=delimiters)
        raw_lines = [line.strip() for line in text.split("\n") if line.strip()]
        reader = clevercsv.reader(raw_lines, dialect=dialect)
        rows = list(reader)
        if processed_cmd.has_header and rows:
            headers_list, rows = rows[0], rows[1:]
        return rows, headers_list


class JsonLinesInput(Input):
    lexer = JsonLexer

    def get_rows(self, text: str, processed_cmd: ProcessedCommand) -> (List[Any], Mapping[int, str]):
        json_rows = [json.loads(line.strip()) for line in text.split("\n") if line.strip()]
        headers_list = []
        seen_headers = set()
        for row in json_rows:
            for k in row:
                if k in seen_headers:
                    continue
                headers_list.append(k)
                seen_headers.add(k)
        rows = []
        for json_row in json_rows:
            row = [json_row.get(h, MISSING) for h in headers_list]
            rows.append(row)
        return rows, headers_list


class ShellInput(Input):
    lexer = TextLexer

    def get_rows(self, text: str, processed_cmd: ProcessedCommand) -> (List[Any], Mapping[int, str]):
        rows = []
        lines = text.split("\n")
        header_line = []
        if processed_cmd.has_header:
            header_line, lines = shlex.split(lines[0], posix=True), lines[1:]
        for line in lines:
            line = line.strip()
            if not line:
                continue
            row = shlex.split(line, posix=True)
            rows.append(row)
        return rows, header_line


class JQInput(Input):
    lexer = JsonLexer

    def __init__(self, args):
        self.args = args

    def get_rows(self, text: str, processed_cmd: ProcessedCommand) -> (List[Any], Mapping[int, str]):
        return run_jq(text, self.args), []


class NopInput(Input):
    lexer = TextLexer

    def get_rows(self, text: str, processed_cmd: ProcessedCommand) -> (List[Any], Mapping[int, str]):
        return text, []


registry = Registry(
    attr="inputs",
    tpe=Input,
    default_alias="c",
    data={
        "c": CSVInput(),
        "jl": JsonLinesInput(),
        "sh": ShellInput(),
        "jq": JQInput,
        "n": NopInput(),
    },
)



from textomatic.exceptions import ProcessException

macros = {}


def as_list(args):
    return [args]


def macro(alias, split=None):
    def wrap(fn):
        macros[alias] = {
            "fn": fn,
            "split": split or as_list,
        }
        return fn

    return wrap


@macro("jq")
def jq_macro(expression=""):
    return f"r;o:jq`{expression}`"


def get(alias):
    try:
        return macros[alias]
    except KeyError:
        raise ProcessException(f"Unregistered macro {alias}")



import sys

import click
from prompt_toolkit.patch_stdout import patch_stdout

from textomatic import context
from textomatic.app.builder import create_app, process_one


@click.command()
@click.argument("path", nargs=-1)
@click.option(
    "-c",
    "--command",
    default="",
    help="Start with an initial COMMAND",
)
@click.option(
    "-p",
    "--process-and-exit",
    is_flag=True,
    help="Run process on INPUT and COMMAND and exit without showing the UI",
)
@click.option(
    "-h",
    "--horizontal",
    is_flag=True,
    help="Start with a horizontal split between the INPUT and OUTPUT",
)
@click.option(
    "-m",
    "--manual",
    is_flag=True,
    help="Start with manual mode enabled. In manual mode, OUTPUT is only updated on demand",
)
@click.option(
    "-f",
    "--focus",
    type=click.Choice(["COMMAND", "INPUT", "OUTPUT", "c", "i", "o"]),
    help="Start the specified component focused",
)
def main(path, command, process_and_exit, horizontal, manual, focus):
    if len(path) > 1:
        raise click.ClickException("Only one path argument is supported")

    if process_and_exit and manual:
        raise click.ClickException("--manual with --process-and-exit makes no sense")

    ctx = context.get()

    path = path[0] if path else None
    if path or not sys.stdin.isatty():
        if path:
            with open(path) as f:
                text = f.read()
        else:
            text = sys.stdin.read()
            sys.stdin.close()
            sys.stdin = open(2)
        ctx.input_buffer.text = text

    ctx.live = not manual
    ctx.box_veritcal_orientation = not horizontal
    ctx.cmd_buffer.text = command

    if process_and_exit:
        ctx.non_interactive = True
        process_one()
    else:
        ctx.app = create_app(focus)
        with patch_stdout():
            ctx.app.run()

    if ctx.print_output_on_exit or ctx.non_interactive:
        print(ctx.output_buffer.text)


if __name__ == "__main__":
    main()



from prompt_toolkit.key_binding import KeyBindings

from textomatic import context

kb = KeyBindings()
cmd_kb = KeyBindings()

kb.add("c-c")(lambda e: e.app.exit())
kb.add("tab")(lambda e: e.app.layout.focus_next())
kb.add("s-tab")(lambda e: e.app.layout.focus_previous())


@kb.add("c-t")
def change_boxes_orientation(_):
    ctx = context.get()
    ctx.box_veritcal_orientation = not ctx.box_veritcal_orientation


@kb.add("c-o")
def exit_and_print_output(_):
    ctx = context.get()
    ctx.print_output_on_exit = True
    ctx.app.exit()


@kb.add("c-p")
def copy_output_to_clipboard(_):
    ctx = context.get()
    ctx.app.clipboard.set_text(ctx.output_buffer.text)
    ctx.copied_to_clipboard = True

    def off():
        ctx.copied_to_clipboard = False

    ctx.app.loop.call_later(0.1, off)


@kb.add("f1")
def toggle_help(_):
    ctx = context.get()
    ctx.display_help = not ctx.display_help


@kb.add("c-l")
def toggle_live(_):
    ctx = context.get()
    ctx.live = not ctx.live


@kb.add("c-r")
def run1(_):
    ctx = context.get()
    ctx.process_fn("run")


cmd_kb.add("c-m")(lambda _: run1(_))



from prompt_toolkit.styles import style_from_pygments_cls, Style, merge_styles
from pygments.styles import get_style_by_name

pygments_style_name = "monokai"
pygments_style = style_from_pygments_cls(get_style_by_name(pygments_style_name))

STATUS_BAR = "status-bar"
STATUS_BAR_ERROR = "status-bar-error"
STATUS_BAR_NOTIFY = "status-bar-notify"
TEXTBOX = "textbox"
TEXTBOX_FOCUSED = "textbox-focused"
COMMAND_FOCUSED = "command-focused"

custom_style = Style(
    [
        ("cursor-line", "bg:#222 nounderline"),
        (STATUS_BAR, "bg:#333"),
        (STATUS_BAR_ERROR, "fg:#fff bg:#8A000D"),
        (STATUS_BAR_NOTIFY, "fg:#fff bg:#3E7259"),
        (TEXTBOX, "bg:#333"),
        (TEXTBOX_FOCUSED, "bg:#60AF8C fg:#000"),
        (COMMAND_FOCUSED, "fg:ansiblue bold"),
        ("frame.label", "fg:#000"),
    ]
)

application_style = merge_styles([pygments_style, custom_style])



import subprocess

from textomatic.exceptions import ProcessException


def run_jq(text, args):
    final_args = ["jq", "-c"]
    if args:
        final_args.append(args)
    text = bytes(text, encoding="utf-8") if text else None
    try:
        return subprocess.check_output(
            final_args,
            stderr=subprocess.STDOUT,
            input=text,
        ).decode()
    except subprocess.CalledProcessError as e:
        raise ProcessException(e.stdout.decode())



from prompt_toolkit.layout import HSplit, Window, BufferControl
from prompt_toolkit.widgets import FormattedTextToolbar

from textomatic import context
from textomatic.app import style


def textbox(buffer, title, lexer=None):
    def get_style():
        ctx = context.get()
        if ctx.app.layout.current_buffer is buffer:
            cls = style.TEXTBOX_FOCUSED
        else:
            cls = style.TEXTBOX
        return f"class:{cls}"

    return HSplit(
        [
            FormattedTextToolbar(
                text=title,
                style=get_style,
            ),
            Window(
                cursorline=True,
                content=BufferControl(
                    buffer=buffer,
                    focus_on_click=True,
                    lexer=lexer,
                ),
                ignore_content_width=True,
            ),
        ]
    )



from contextvars import ContextVar
from dataclasses import dataclass

from prompt_toolkit import Application
from prompt_toolkit.buffer import Buffer

from textomatic.model import ProcessedCommand, ProcessedInput


@dataclass
class AppContext:
    app: Application = None
    process_fn = None
    box_veritcal_orientation = True
    display_help = False
    print_output_on_exit: bool = False
    non_interactive = False
    live = True

    current_error: str = None
    copied_to_clipboard = False
    dirty = False

    cmd_buffer: Buffer = Buffer(multiline=False)
    input_buffer: Buffer = Buffer()
    output_buffer: Buffer = Buffer(read_only=True)


@dataclass
class ProcessContext:
    processed_command: ProcessedCommand = ProcessedCommand("")
    processed_input: ProcessedInput = None


_current = ContextVar("current", default=AppContext())
_current_process = ContextVar("current_process", default=ProcessContext())


def get() -> AppContext:
    return _current.get()


def get_process() -> ProcessContext:
    return _current_process.get()


def reset():
    _current.set(AppContext())


def reset_process():
    _current_process.set(ProcessContext())



from dataclasses import dataclass, field
from typing import Any, List


@dataclass
class ProcessedCommand:
    cmd: str
    delimiter: str = None
    outputs: List[Any] = field(default_factory=list)
    inputs: List[Any] = field(default_factory=list)
    structure: object = None
    types: object = None
    has_header: bool = False
    headers: dict = None
    raw: bool = False


@dataclass
class ProcessedInput:
    rows: list
    headers: dict


@dataclass
class NotSet:
    pass


@dataclass
class Missing:
    pass


NO_DEFAULT = NotSet()
MISSING = Missing()



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['textomatic', 'textomatic.app', 'textomatic.processor']

package_data = \
{'': ['*']}

install_requires = \
['clevercsv>=0.6.4,<0.7.0',
 'click>=7.1.2,<8.0.0',
 'prompt-toolkit>=3.0.7,<4.0.0',
 'pygments>=2.7.1,<3.0.0',
 'pyparsing>=2.4.7,<3.0.0',
 'pyperclip>=1.8.0,<2.0.0',
 'tabulate>=0.8.7,<0.9.0']

entry_points = \
{'console_scripts': ['tm = textomatic.main:main']}

setup_kwargs = {
    'name': 'textomatic',
    'version': '0.3.0',
    'description': 'Scratchpad for tabular data transformations, and then some',
    'long_description': '# textomatic\nScratchpad for tabular data transformations\n\nGiven input of some tabular data (AKA `INPUT`), \nand a command (AKA `COMMAND`), transform that input \ninto some other output (AKA `OUTPUT`).\n\n![Screencast Demo](https://i.ibb.co/bL9MJkm/demo.gif)\n\n## Installation\n```shell script\npip install textomatic\n```\n\nAt the moment, Python 3.8 or greater is required.\n\n## Running\nStart it by running\n```shell script\n$ tm\n```\n\nThe above will start `textomatic` with a blank slate.\n\nYou can load a file by passing it as the first argument:\n```shell script\n$ tm <PATH_TO_FILE>\n```\n\nYou can also pipe content from stdin:\n```shell script\n$ ls | tm\n```\n\nTo see what arguments/options are available, run:\n```\n$ tm --help\n```\n\nWhen `textomatic` is running, type `F1` to see available keyboard shortcuts.\n\nTo exit `textomatic` use `CTRL-C` to exit without any output or `CTRL-O` to print\ncurrent `OUTPUT` into standard out.\n\nUse `CTRL-P` to put the current `OUTPUT` in the system clipboard.\n\n## Examples\n\n### Parsing `ps aux` output\nPipeing content from shell using `ps aux | tm` and transforming it into json lines where\neach line containes the `USER` and `PID` columns with lower cased keys.\n`PID` is casted into an integer.\n\n`COMMAND`:\n* `h` says `INPUT` includes a header\n* `i:sh` says the `INPUT` should be parsed like shell output\n* `s:{user:USER,pid:PID}` specifies the output structure\n* `o:jl` specifies the output format to be json lines\n* `t:PID:i` specifies the `PID` column shouldd ge parsed as integer\n```\nINPUT                                  │OUTPUT                                  \nUSER               PID  %CPU %MEM      │{"user": "dan", "pid": 63507}           \ndan              63507   6.3  0.4  6178│{"user": "_windowserver", "pid": 250}   \n_windowserver      250   4.4  0.3 12494│{"user": "dan", "pid": 54987}           \ndan              54987   3.8 12.8 16080│{"user": "_hidd", "pid": 184}           \n_hidd              184   2.0  0.0  5608│                                        \n                                       │                                        \n                                       │                                        \n                                       │                                        \n                                       │                                        \n                                       │                                        \n                                       │                                        \n                                       │                                        \n                                       │                                        \n                                       │                                        \n                                       │                                        \n                                       │                                        \n                                       │                                        \n                                       │                                        \n                                       │                                        \n                                       │                                        \n                                       │                                        \nvi-insert|COMMAND|live|in:sh|out:jl|delim:auto|header:true                      \n> h;i:sh;s:{user:USER,pid:PID};o:jl;t:PID:i                                     \n```\n\n### Pretty printing csv\n`COMMAND`:\n* `h` says `INPUT` includes a header\n* `o:t` specifies the output format to be a pretty printed table\n```\nINPUT                                  │OUTPUT                                  \nName,Age,City                          │╒═══════════╤═══════╤════════╕          \nJames Joe,34,NYC                       ││ Name      │   Age │ City   │          \nJohn Doe,25,London                     │╞═══════════╪═══════╪════════╡          \n                                       ││ James Joe │    34 │ NYC    │          \n                                       │├───────────┼───────┼────────┤          \n                                       ││ John Doe  │    25 │ London │          \n                                       │╘═══════════╧═══════╧════════╛          \n                                       │                                        \n                                       │                                        \n                                       │                                        \n                                       │                                        \n                                       │                                        \n                                       │                                        \n                                       │                                        \n                                       │                                        \n                                       │                                        \n                                       │                                        \n                                       │                                        \n                                       │                                        \n                                       │                                        \n                                       │                                        \nvi-insert|COMMAND|live|in:c|out:t|delim:auto|header:true                        \n> h;o:t                                                                         \n```\n\n### Correcting bad input\n`COMMAND`:\n* `h` says `INPUT` includes a header\n* `t:Age:i/76/` provides a default value when the `Age` column cannot be parsed as integer\n```\nINPUT                                  │OUTPUT                                  \nName,Age,City                          │[[\'James Joe\', 34, \'NYC\'],              \nJames Joe,34,NYC                       │ [\'John Doe\', 76, \'London\'],            \nJohn Doe,Not a number,London           │ [\'Jane Row\', 24, \'Tel Aviv\']]          \nJane Row,24,Tel Aviv                   │                                        \n                                       │                                        \n                                       │                                        \n                                       │                                        \n                                       │                                        \n                                       │                                        \n                                       │                                        \n                                       │                                        \n                                       │                                        \n                                       │                                        \n                                       │                                        \n                                       │                                        \n                                       │                                        \n                                       │                                        \n                                       │                                        \n                                       │                                        \n                                       │                                        \n                                       │                                        \nvi-insert|COMMAND|live|in:c|out:l|delim:auto|header:true                        \n> h;t:Age:i/76/                                                                 \n```\n\n### Interactive `jq`\n`COMMAND`:\n* `r` puts evaluation into raw mode where input is not assumed to be raw based\n* `o:jq` use the `jq` output and pass arguments to it wrapped with backticks.\n```\nINPUT                                  │OUTPUT                                  \n{"one": 1, "two": 2, "three": 3}       │{"o":1,"t1":2,"t2":3}                   \n{"one": 1.0, "two": 2.0, "three": 3.0} │{"o":1,"t1":2,"t2":3}                   \n                                       │                                        \n                                       │                                        \n                                       │                                        \n                                       │                                        \n                                       │                                        \n                                       │                                        \n                                       │                                        \n                                       │                                        \n                                       │                                        \n                                       │                                        \n                                       │                                        \n                                       │                                        \n                                       │                                        \n                                       │                                        \n                                       │                                        \n                                       │                                        \n                                       │                                        \n                                       │                                        \n                                       │                                        \nvi-insert|COMMAND|live|in:c|out:jq|delim:auto|header:false|raw:true             \n> r;o:jq`{o: .one, t1: .two, t2: .three}`                                       \n```\n\n## Usage\n`textomatic` is split into 3 parts:\n* `INPUT`: The input data that is to be transformed\n* `OUTPUT`: The result of applying the `COMMAND` on `INPUT`\n* `COMMAND`: The transformation logic using a succinct expression language described below\n    \nUse the `Tab` key to move between them.\n\n`INPUT` and `COMMAND` are edited using `vim` bindings.\n\n### `COMMAND` Expression Language\n`COMMAND` is composed of different parts separated by `;`, e.g.\n\n```\n> h;i:c;o:jl;s:[1,2]\n```\n\nThe above will be explained in detail later on but for now we can see it has 4 parts:\n* `h`: specifies that the input `csv` has a header\n* `i:c` specifies that the input is in fact a `csv`\n* `o:jl` specified that the output should be in jsonlines format\n* `s:[1,2]` specifies that the output should only include the first and\n            second columns from the input, in that order\n\nTo use a different expression separator, start the command with `:<SEP>`, e.g.\n```\n> :|h|i:c\n```\n\n#### Expressions\n\n##### The `h` expression (header)\nThe simplest expression. It is basically a flag denoting whether the input\nincludes headers. This is relevent for inputs like `csv`\n\n##### The `d` expression (delimiter)\nUsed by the `csv` input. In most cases, the delimiter can be automatically\ndeduced. In cases where it cannot, use `d`, e.g. to set a `,` delimiter:\n```\n> d:,\n```\n\nTo specify delimiters that are not easy to input, start the delimiter with\n`\\`, the remaining part will then be parsed as a python literal\n(wrapped in string), for example this will set the delimiter to the unicode\ncharacter `─` (U+2500):\n```\n> d:\\u2500\n```\n\n##### The `i` expression (input)\nUsed to specify the input format. Currently these are the available inputs:\n* `c` (`csv`, this is the default. The delimiter musn\'t be a `,` as the name may imply)\n* `jl` (jsonlines)\n* `sh` (shell, e.g. the output of `ps aux`)\n* `jq` (Using [jq](https://stedolan.github.io/jq/))\n\n##### The `o` expression (output)\nUsed to specify the output format. Currently these are the available outputs:\n* `l` (python literal, this is the default)\n* `j` (json)\n* `jl` (jsonlines)\n* `c` (csv)\n* `t` (pretty printed table)\n* `h` (table html)\n* `jq` (Using [jq](https://stedolan.github.io/jq/))\n\n##### The `t` expression (types)\nFor inputs with no clear types (e.g. csv/shell), all columns are initially assumed to\nbe strings. To modify types of different columns use `t`.\n\nThe types are:\n* `s` (string, the default)\n* `f` (float)\n* `i` (integer)\n* `b` (boolean, case insensitive `true/yes/y/on/1` will be parsed as `true`)\n* `j` (json, will JSON parse the column)\n* `l` (literal, will parse a python literal using `ast.literal_eval`)\n\nUsing positional syntax:\n```\n> t:i,i,b\n```\nIn the above:\n* the first 2 columns will be parsed as integers\n* the third column will be parsed as boolean\n* the rest will be strings\n\nUsing indexed syntax:\n```\n> t:1:i,3:f\n```\nIn the above:\n* the first column will be parsed as integer\n* the third column will be parsed as float\n* the rest will be strings\n\nUsing named syntax:\n```\n> t:col1:b,col4:i\n```\nIn the above, assuming the input contains headers:\n* the column named `col1` will be parsed as boolean\n* the column named `col4` will be parsed as integer\n* the rest will be strings\n\nOptional types:\n\nIf a certain value may be invalid, you can use `?` to\nmark it as optional, in this case, its value will be \nconverted to `null` when it is invalid:\n```\n> t:col1:i?,col4:i?\n```\n\nDefaults:\n\nIf you want the specify a value different than `null` for invalid entries\nuse the following syntax:\n```\n> t:col1:i/0/,col4:f/0.0/\n```\n\nThe value between the `/`\'s will be evaluated as a python literal.\n\n##### The `s` expression (structure)\nThe `s` expression is used to specify the structure of the `OUTPUT`.\nSome of the options are catered to the python output but they will fallback\nto a reasonable alternative for other outputs.\n\nSimple transformations:\n* `s:[]` - Each row will be a list in `OUTPUT`\n* `s:()` - Each row will be a tuple in `OUTPUT`\n* `s:{}` - Each row will be an object in `OUTPUT` (assumes input has headers)\n* `s:d()` - Same as `s:{}`\n* `s:s()` - Each row will be as set in `OUTPUT`\n\nComplex transformations:\n* `s:[1,2,col6]` - Each row will contain the first and second columns and a column\n                   named `col6`. Note that it is wrapped with `[]`. This only\n                   means the output row will be a list. You can just as well\n                   wrap it with `{}` to get objects e.g. `s:{1,2,col6}`.\n                   The different wrapping options are desribed in "Simple Transformations"\n                   of the previous section.\n* `s:{first:1,second:2}` - Each row will contain the first and second columns with\n                           The specified new headers (`first` and `second`)\n* `s:[-2,-1]` - Each row will contain the two last columns from `INPUT`\n* `s:{1,second:2}` - Column definitions can be mixed.\n\nNested transformations:\n* `s:[{1,2},{3,4}]` - Each row will contain two objects, the first object will\n                      contain the first and second columns, the second object\n                      will contain the third and fourth columns.\n\nNesting can be as complex as you wish and rules from previous sections\ncan be applied freely. As an overly complex example:\n```\n> s:{k1:1,k2:{2,three},k3:[{four,5},d(-4,s7:seven)],eight,k4:(one, two, (four, five)),s:s(1,1,1)}\n``` \n\nFetching values of nested data:\n* `s:[some_obj.key1.key2]` - Each row will contain a single nested value from\n                             the `some_obj` column. \n* `s:[some_obj.key1?.key2]` - Same as above, but don\'t fail if `key1` doesn\'t exist,\n                              instead, replace it with `null`.\n* `s:[some_obj.key1?.key2/100/]` - Same as above, but use `100` instead of `null`.\n* `s:[some_obj.key1.key2/100/]` - Default value without optional `?` is equivalent to\n                                  `s:[some_obj?.key1?.key2?/100/]`                               \n\nA caveat of using `//` for default values is that the default value cannot\ninclude `/`. This is mostly due to a very simple parser that is currently\nimplemented for the expression language.',
    'author': 'Dan Kilman',
    'author_email': 'dankilman@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/dankilman/textomatic',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



import string
from dataclasses import dataclass
from typing import List, Union, Any

from pyparsing import (
    Literal,
    ZeroOrMore,
    Empty,
    Word,
    Optional,
    Combine,
    printables,
    nums,
    QuotedString,
    Suppress,
    Forward,
    alphanums,
)

from textomatic.exceptions import ProcessException
from textomatic.model import NO_DEFAULT


class ParseData:
    pass


@dataclass
class KeyToValueData(ParseData):
    key: Any
    value: Any


@dataclass
class TypeDefData(ParseData):
    type: str
    optional: bool
    default: Any


@dataclass
class StructureData(ParseData):
    type: str
    fields: object


@dataclass
class RefData(ParseData):
    value: list
    default: Any


@dataclass
class LocData(ParseData):
    value: object
    optional: bool = False


@dataclass
class IdData(ParseData):
    value: object
    optional: bool = False


@dataclass
class ProcessorData(ParseData):
    alias: str
    args: str


# general
OptionalMarker = Optional("?")("optional")
PrintablesReducedForDefault = Word(string.printable, excludeChars="/")
PrintablesReducedForArgs = Word(string.printable, excludeChars="`")
Default = Optional(Combine(Suppress("/") + PrintablesReducedForDefault + Suppress("/")))("default")

# ref
Loc = Combine(Optional("-") + Word(nums)) + OptionalMarker
Loc.setParseAction(lambda t: [LocData(int(t[0]), optional=bool(t.optional))])
PrintablesReducedForId = Word(printables, excludeChars="}]),./:?")
FreeFormForId = QuotedString("'") | QuotedString('"') | PrintablesReducedForId
Id = FreeFormForId + OptionalMarker
Id.setParseAction(lambda t: [IdData(t[0], optional=bool(t.optional))])
Ref = Loc | Id
RefPath = (Ref + ZeroOrMore(Suppress(".") + Ref))("ref_path") + Default

# types
AnyType = Word("sifbjld_", exact=1) + OptionalMarker + Default
AnyType.setParseAction(lambda t: [TypeDefData(t[0], optional=bool(t.optional), default=t.default or NO_DEFAULT)])
DefaultType = Empty()
DefaultType.setParseAction(lambda t: [TypeDefData("_", optional=False, default=NO_DEFAULT)])
RefToType = Ref + Suppress(":") + AnyType
RefToType.setParseAction(lambda t: [KeyToValueData(t[0], t[1])])
TypeDef = RefToType | AnyType | DefaultType
Types = (TypeDef + ZeroOrMore(Suppress(",") + TypeDef))("types")
Types.setParseAction(lambda t: [t.types.asList()])

# structure
Structure = Forward()
StructureRef = RefPath.copy()
StructureRef.setParseAction(lambda t: [RefData(t.ref_path.asList(), default=t.default or NO_DEFAULT)])
StructureOrRef = Structure | StructureRef
IdToStructureOrRef = Id + Suppress(":") + StructureOrRef
IdToStructureOrRef.setParseAction(lambda t: [KeyToValueData(t[0], t[1])])
Field = IdToStructureOrRef | StructureOrRef
Fields = (Field + ZeroOrMore(Suppress(",") + Field))("fields")
Fields.setParseAction(lambda t: [t.fields.asList()])
Structure << ("[" + Fields + "]" | "{" + Fields + "}" | "(" + Fields + ")" | "d(" + Fields + ")" | "s(" + Fields + ")")
Structure.setParseAction(lambda t: [StructureData(t[0] + t[-1], t[1])])
EmptyStructure = Literal("[]") | "{}" | "()" | "d()" | "s()"
EmptyStructure.setParseAction(lambda t: [StructureData(t[0], [])])
TopLevelStructure = EmptyStructure | Structure

# processor (inputs/outputs)
ProcessorArgsInner = Optional(PrintablesReducedForArgs)
ProcessorArgs = (Suppress("`") + ProcessorArgsInner + Suppress("`"))("args")
ProcessorAlias = Word(alphanums)("alias")
Processor = ProcessorAlias + Optional(ProcessorArgs)
Processor.setParseAction(lambda t: [ProcessorData(t.alias, (t.args or [""])[0])])
Processors = (Processor + ZeroOrMore(Suppress(",") + Processor))("processors")
Processors.setParseAction(lambda t: [t.processors.asList()])


# api
def parse_types(expr) -> List[Union[KeyToValueData, TypeDefData]]:
    return _parse(Types, expr)


def parse_structure(expr) -> StructureData:
    return _parse(TopLevelStructure, expr)


def parse_processors(expr) -> ProcessorData:
    return _parse(Processors, expr)


# internal
def _parse(element, expr):
    try:
        parsed = element.parseString(expr, parseAll=True)
    except Exception:
        raise ProcessException(f"Not a valid expresion: {expr}")
    return parsed.asList()[0]



import string
import unicodedata


def as_printable(s):
    if not s:
        return s
    r = []
    for c in s:
        if c in string.printable:
            r.append(c)
        elif unicodedata.category(c).startswith("C"):
            r.append("?")
        else:
            r.append(c)
    return "".join(r)
